from django import template
from django.urls import reverse

register = template.Library()


@register.simple_tag(takes_context=True)
def crud_link(context, url_name, pk_param=None, pk=None):
    kwargs = {}
    if pk_param and pk:
        kwargs[pk_param] = pk

    for key, val in context['url'].__dict__.items():
        if key.startswith('_'):
            continue

        if context['crud']['pk_param'] == key:
            continue

        kwargs[key] = val

    return reverse(url_name, kwargs=kwargs) + context['crud']['link_suffix']


@register.simple_tag(takes_context=True)
def crud_field(context, object, field):
    return getattr(object, field)



import json
from pprint import pprint

from django.http import HttpResponse
from django.utils.decorators import classonlymethod
from django.views.decorators.csrf import ensure_csrf_cookie, csrf_exempt
from django.views.generic.base import View, ContextMixin, TemplateResponseMixin
from zmei.json import ZmeiReactJsonEncoder


class ImproperlyConfigured(Exception):
    pass


class _Data(object):
    def __init__(self, data=None):
        self.__dict__.update(data or {})

    def __add__(self, data):
        return _Data({**self.__dict__, **data})


class ZmeiDataViewMixin(ContextMixin, View):
    _data = None

    def get_data(self, url, request, inherited=False):
        return {}

    def _get_data(self):
        if not self._data:
            self._data = self.get_data(
                url=type('url', (object,), self.kwargs),
                request=self.request,
                inherited=False
            )

        return self._data

    def get_context_data(self, **kwargs):
        context_data = super().get_context_data(**kwargs)

        return {**context_data, **self._get_data()}


class ZmeiRemoteInvocationViewMixin(ZmeiDataViewMixin):

    def get(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)

        if 'application/json' in self.request.META['HTTP_ACCEPT']:
            return HttpResponse(ZmeiReactJsonEncoder(view=self).encode(self._get_data()), content_type='application/json')

        return self.render_to_response(context)

    @classonlymethod
    def as_view(cls, **initkwargs):
        return ensure_csrf_cookie(super().as_view(**initkwargs))

    def _remote_response(self, data):
        return HttpResponse(ZmeiReactJsonEncoder(view=self).encode(data), content_type='application/json')

    def state(self, data):
        return {'__state__': data}

    def error(self, data):
        return {'__error__': data}

    def post(self, request, *args, **kwargs):
        if 'application/json' not in self.request.META['HTTP_ACCEPT']:
            raise ValueError('Only json is available as a response type.')

        call = json.loads(request.body)
        method_name = f"_remote__{call.get('method')}"

        if not hasattr(self, method_name):
            raise ValueError('Unknown method')

        method = getattr(self, method_name)

        try:
            result = method(
                type('url', (object,), self.kwargs),
                request,
                *(call.get('args') or [])
            )
        except Exception as e:
            return self._remote_response(self.error(str(e)))

        if not result:
            return self._remote_response(self.state(self._get_data()))

        return self._remote_response(result)


class CrudView(TemplateResponseMixin):
    def render_to_response(self, context, **response_kwargs):
        return context


class CrudMultiplexerView(TemplateResponseMixin, ContextMixin, View):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.crud_views = {}
        for cls in self.get_crud_views():
            crud = cls(*args, **kwargs)
            self.crud_views[crud.name] = crud

    def get_crud_views(self):
        return ()

    def get(self, request, *args, **kwargs):
        context = self.get_context_data(**kwargs)

        context['crud'] = {}
        for crud in self.crud_views.values():
            self.populate_crud(args, crud, kwargs, request)

            context['crud'][crud.name] = crud.get(request, *args, **kwargs)

        return self.render_to_response(context)

    def populate_crud(self, args, crud, kwargs, request):
        crud.request = request
        crud.args = args
        crud.kwargs = kwargs

    def post(self, request, *args, **kwargs):
        form_name = request.POST.get('_form')
        crud = self.crud_views.get(form_name)

        self.populate_crud(args, crud, kwargs, request)

        if not crud:
            return self.http_method_not_allowed(request, *args, **kwargs)

        return crud.post(request, *args, **kwargs)



from django.db.models import QuerySet, Model
from django.utils.module_loading import import_string
from rest_framework import serializers


class DefaultModelSerializer(serializers.ModelSerializer):
    class Meta:
        model = None
        exclude = []


def create_default_serializer(model):
    serializer = type('_', (DefaultModelSerializer,), {})
    serializer.Meta.model = model

    return serializer


def get_model_serializer(model, descriptor='_', strict=False):
    serializer = None
    try:
        index_import_path = f"{'.'.join(model.__module__.split('.')[:-1])}.serializers.index"
        serializer_index = import_string(index_import_path)

        if model in serializer_index:
            serializer = serializer_index[model].get(descriptor)
            if not serializer and strict:
                raise AttributeError(f'Serializer with name "{descriptor}" does not exist for model "{model}"')
    except ImportError:
        pass

    if not serializer:
        serializer = create_default_serializer(model)

    return serializer


def serialize(qs_or_model, descriptor='_'):
    if isinstance(qs_or_model, QuerySet):
        model = qs_or_model.model
        many = True
    elif isinstance(qs_or_model, Model):
        model = qs_or_model
        many = False
    else:
        raise AttributeError('serialize method accept only QuerySet or Model. Given: ', type(qs_or_model))

    return get_model_serializer(model, descriptor=descriptor, strict=True)(qs_or_model, many=many).data


import json
import os

from django.conf import settings
from django.http import HttpResponse
from py_mini_racer import py_mini_racer
from py_mini_racer.py_mini_racer import MiniRacerBaseException
from zmei.json import ZmeiReactJsonEncoder

from .views import ZmeiDataViewMixin, ImproperlyConfigured, ZmeiRemoteInvocationViewMixin


class ZmeiReactServer(object):
    def __init__(self):
        super().__init__()

        self.loaded_files = []
        self.loaded_files_mtime = {}

        self.jsi = None

        self.checksum = None

    def reload_interpreter(self):
        self.jsi = py_mini_racer.MiniRacer()

        code = """
        var global = this;
        var module = {exports: {}};
        var setTimeout = function(){};
        var clearTimeout = function(){};var console = {
            error: function() {},
            log: function() {},
            warn: function() {}
        };
        """

        self.jsi.eval(code)

        for filename in self.loaded_files:
            self.loaded_files_mtime[filename] = os.path.getmtime(filename)
            self.eval_file(filename)

    def autreload(self):
        if len(self.loaded_files_mtime) == 0:
            return

        for filename in self.loaded_files:
            if self.loaded_files_mtime[filename] != os.path.getmtime(filename):
                print('Reloading ZmeiReactServer')
                self.reload_interpreter()
                break

    def evaljs(self, code):
        if not self.jsi:
            self.reload_interpreter()

        return self.jsi.eval(code)

        # except JSRuntimeError as e:
        #     message = str(e)
        #
        #     message = '\n' + colored('Error:', 'white', 'on_red') + ' ' + message
        #
        #     print(message)
        #     m = re.search('\(line\s+([0-9]+)\)', message)
        #     if m:
        #         print('-' * 100)
        #         print('Source code:')
        #         print('-' * 100)
        #         row = int(m.group(1)) - 1
        #         source = code.splitlines()
        #
        #         line = colored(source[row], 'white', 'on_red')
        #         print('\n'.join([f'{x+1}:\t{source[x]}' for x in range(max(0, row - 10), row)]))
        #         print(f'{row+1}:\t{line}')
        #         print('\n'.join([f'{x+1}:\t{source[x]}' for x in range(row + 1, min(row + 10, len(source) - 1))]))
        #         print('-' * 100)

    def load(self, filename):
        self.loaded_files.append(filename)

    def eval_file(self, filename):
        with open(filename) as f:
            self.evaljs(f.read())


class ZmeiReactViewMixin(ZmeiRemoteInvocationViewMixin):
    react_server = None
    react_components = None
    server_render = True

    def get_context_data(self, **kwargs):
        data = super().get_context_data(**kwargs)

        if not isinstance(self.react_server, ZmeiReactServer):
            raise ImproperlyConfigured('ZmeiReactViewMixin requires react_server property')

        if not isinstance(self.react_components, list):
            raise ImproperlyConfigured('ZmeiReactViewMixin requires react_component property')

        data['react_state'] = ZmeiReactJsonEncoder(view=self).encode(self._get_data())

        if settings.DEBUG:
            self.react_server.autreload()

        if self.server_render:
            for cmp in self.react_components:
                try:
                    data[f'react_page_{cmp}'] = self.react_server.evaljs(f"R.renderServer(R.{cmp}Reducer, R.{cmp}, {data['react_state']});")

                    # print('WARN! Server-side rendering disabled!')
                except MiniRacerBaseException as e:
                    data[f'react_page_{cmp}'] = f'<script>var err = {json.dumps({"msg": str(e)})}; ' \
                                                f'document.body.innerHTML = ' \
                                                "'<h2>Error rendering React component. See console for details.</h2>' + " \
                                                f'"<pre>" + err.msg + "</pre>" + document.body.innerHTML;</script>'

        return data
